/**
  *  \file planetscentral/talk/syntaxscript.cc
  *  \brief Talk: Syntax Highlighter for CCScript
  */

#include "planetscentral/talk/syntaxscript.h"
#include "int/keywords.h"

namespace {
    enum Class {
        cSpace    = 1,
        cNewline  = 2,
        cComment  = 4,
        cLParen   = 8,
        cRParen   = 16,
        cDot      = 32,
        cQuote    = 64,
        cComma    = 128,
        cDigit    = 256,
        cLetter   = 512,
        cOther    = 1024
    };

    Class classify(char c) {
        switch (c) {
         case ' ':
         case '\t':
         case '\v':
         case '\f':
            return cSpace;

         case '\r':
         case '\n':
            return cNewline;

         case '%':
            return cComment;

         case '(':
            return cLParen;

         case ')':
            return cRParen;

         case '.':
            return cDot;

         case '\'':
         case '"':
            return cQuote;

         case ',':
            return cComma;

         default:
            if (c >= '0' && c <= '9') {
                return cDigit;
            } else if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '_') {
                return cLetter;
            } else {
                return cOther;
            }
        }
    }

    string_t::size_type skip(const string_t& str, string_t::size_type& pos, int c)
    {
        string_t::size_type n = pos;
        while (n < str.size() && (classify(str[n]) & c) != 0) {
            ++n;
        }

        string_t::size_type result = n - pos;
        pos = n;
        return result;
    }

    bool skip1(const string_t& str, string_t::size_type& pos, int c)
    {
        bool result = false;
        if (pos < str.size() && (classify(str[pos]) & c) != 0) {
            result = true;
            ++pos;
        }
        return result;
    }

    void skipString(const string_t& str, string_t::size_type& pos)
    {
        const char delim = str[pos-1];
        bool quote = false;
        while (pos < str.size()) {
            const char ch = str[pos++];
            if (quote) {
                quote = false;
            } else if (delim == '"' && ch == '\\') {
                quote = true;
            } else if (ch == delim) {
                break;
            } else {
                // normal character
            }
        }
    }
}


SyntaxScript::SyntaxScript(const SyntaxDatabase& db)
    : db(db),
      str(),
      pos(0),
      state(sDefaultBOL),
      paren_level(0)
{ }

SyntaxScript::~SyntaxScript()
{ }

void
SyntaxScript::init(const string_t& str)
{
    this->str = str;
    this->pos = 0;
    this->state = sDefaultBOL;
    this->paren_level = 0;
}

bool
SyntaxScript::scan(Segment& result)
{
    result.start(pos);
    if (skip(str, pos, cSpace)) {
        // Whitespace
        result.finish(DefaultFormat, pos);
        return true;
    } else if (skip(str, pos, cNewline)) {
        // Newline
        state = sDefaultBOL;
        paren_level = 0;
        result.finish(DefaultFormat, pos);
        return true;
    } else if (skip(str, pos, cComment)) {
        // Comment
        skip(str, pos, ~cNewline);
        result.finish(CommentFormat, pos);
        return true;
    } else if (skip1(str, pos, cLParen)) {
        // Left paren
        ++paren_level;
        result.finish(DefaultFormat, pos);
        leaveDefault();
        return true;
    } else if (skip1(str, pos, cRParen)) {
        // Right paren
        if (paren_level) {
            --paren_level;
        }
        result.finish(DefaultFormat, pos);
        leaveDefault();
        return true;
    } else if (skip1(str, pos, cQuote)) {
        // String
        skipString(str, pos);
        result.finish(StringFormat, pos);
        leaveDefault();
        return true;
    } else if (skip(str, pos, cComma)) {
        // Comma
        leaveDefault();
        if (state == sAfterSubDef && paren_level == 1) {
            state = sAfterSub;
        } else if (state == sAfterDimDef && paren_level == 0) {
            state = sAfterDim;
        } else {
            // nothing particular
        }
        result.finish(DefaultFormat, pos);
        return true;
    } else if (skip(str, pos, cDigit)) {
        // Number
        skip(str, pos, cDigit | cDot);
        result.finish(DefaultFormat, pos);
        leaveDefault();
        return true;
    } else if (skip(str, pos, cLetter)) {
        // Identifier
        skip(str, pos, cLetter | cDot | cDigit);
        string_t id(strUCase(string_t(str, result.start_pos, pos - result.start_pos)));
        IntKeyword kw = lookupKeyword(id);

        // The following keywords are not reported by lookupKeyword.
        // Translate them into a harmless keyword that doesn't cause unnecessary state transitions.
        if (kw == kwNone && (id == "TRUE" || id == "FALSE" || id == "PI" || id == "AND" || id == "OR"
                             || id == "XOR" || id == "NOT" || id == "MOD"))
        {
            kw = kwPrint;
        }

        // Perform state transitions
        switch (state) {
         case sDefaultBOL:
            if (kw == kwNone) {
                // Not a keyword
                state = sDefault;
                result.finish(DefaultFormat, pos);
            } else if (kw == kwSub || kw == kwFunction) {
                // Sub/Function: handle that it is followed by a name, and a parameter list
                state = sAfterSub;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwDim || kw == kwLocal || kw == kwStatic || kw == kwShared || kw == kwCreatePlanetProperty || kw == kwCreateShipProperty) {
                // Dim/Local/Static/Shared: handle that it is followed by an initializer list
                state = sAfterDim;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwFor) {
                // For: "To" is a keyword, "Do" ends the statement
                state = sAfterFor;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwDo || kw == kwLoop) {
                // Do/Loop: "Until" is a keyword
                state = sAfterLoop;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwIf) {
                // If: "Then" is a keyword and ends the statement
                state = sAfterIf;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwCase) {
                // Case: "Is" is a keyword
                state = sAfterCase;
                result.finish(KeywordFormat, pos);
            } else if (kw == kwOn || kw == kwWith) {
                // On/With: "Do" ends the statement
                state = sAfterWith;
                result.finish(KeywordFormat, pos);
            } else {
                // Just a keyword
                state = sDefault;
                result.finish(KeywordFormat, pos);
            }
            break;

         case sDefault:
         case sAfterSubDef:
         case sAfterDimDef:
            result.finish(kw == kwNone ? DefaultFormat : KeywordFormat, pos);
            break;

         case sAfterSub:
            if (kw == kwNone && id != "OPTIONAL") {
                if (paren_level == 0) {
                    result.finish(NameFormat, pos);
                } else if (paren_level == 1) {
                    result.finish(NameFormat, pos);
                    state = sAfterSubDef;
                } else {
                    result.finish(DefaultFormat, pos);
                }
            } else {
                result.finish(KeywordFormat, pos);
            }
            break;
            
         case sAfterDim:
            if (kw == kwNone) {
                if (paren_level == 0) {
                    result.finish(NameFormat, pos);
                    state = sAfterDimDef;
                } else {
                    result.finish(DefaultFormat, pos);
                }
            } else {
                result.finish(KeywordFormat, pos);
            }
            break;

         case sAfterFor:
            result.finish(kw == kwNone && id != "TO" ? DefaultFormat : KeywordFormat, pos);
            if (kw == kwDo) {
                state = sDefaultBOL;
            }
            break;

         case sAfterLoop:
            result.finish(kw == kwNone && id != "UNTIL" ? DefaultFormat : KeywordFormat, pos);
            break;

         case sAfterIf:
            result.finish(kw == kwNone && id != "THEN" ? DefaultFormat : KeywordFormat, pos);
            if (id == "THEN") {
                state = sDefaultBOL;
            }
            break;

         case sAfterCase:
            result.finish(kw == kwNone && id != "IS" ? DefaultFormat : KeywordFormat, pos);
            break;

         case sAfterWith:
            result.finish(kw == kwNone ? DefaultFormat : KeywordFormat, pos);
            if (kw == kwDo) {
                state = sDefaultBOL;
            }
            break;
        }
        return true;
    } else if (skip(str, pos, cDot | cOther)) {
        // Anything else
        result.finish(DefaultFormat, pos);
        leaveDefault();
        return true;
    } else {
        return false;
    }
}

/** Leave default state. Call whenever a non-whitespace token is consumed.
    This will reset the state from sDefaultBOL to sDefault, to turn off statement recognition. */
void
SyntaxScript::leaveDefault()
{
    if (state == sDefaultBOL) {
        state = sDefault;
    }
}
